"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest

from cvxpy import Constant, Variable, log
from cvxpy.settings import QUASILINEAR, UNKNOWN


class TestCurvature(unittest.TestCase):
    """ Unit tests for the expression/curvature class. """

    def setUp(self) -> None:
        self.cvx = Variable()**2
        self.ccv = Variable()**0.5
        self.aff = Variable()
        self.const = Constant(5)
        self.unknown_curv = log(Variable()**3)

        self.pos = Constant(1)
        self.neg = Constant(-1)
        self.zero = Constant(0)
        self.unknown_sign = self.pos + self.neg

    def test_add(self) -> None:
        self.assertEqual((self.const + self.cvx).curvature, self.cvx.curvature)
        self.assertEqual((self.unknown_curv + self.ccv).curvature, UNKNOWN)
        self.assertEqual((self.cvx + self.ccv).curvature, UNKNOWN)
        self.assertEqual((self.cvx + self.cvx).curvature, self.cvx.curvature)
        self.assertEqual((self.aff + self.ccv).curvature, self.ccv.curvature)

    def test_sub(self) -> None:
        self.assertEqual((self.const - self.cvx).curvature, self.ccv.curvature)
        self.assertEqual((self.unknown_curv - self.ccv).curvature, UNKNOWN)
        self.assertEqual((self.cvx - self.ccv).curvature, self.cvx.curvature)
        self.assertEqual((self.cvx - self.cvx).curvature, UNKNOWN)
        self.assertEqual((self.aff - self.ccv).curvature, self.cvx.curvature)

    def test_sign_mult(self) -> None:
        self.assertEqual((self.zero * self.cvx).curvature, self.aff.curvature)
        self.assertEqual((self.neg*self.cvx).curvature, self.ccv.curvature)
        self.assertEqual((self.neg*self.ccv).curvature, self.cvx.curvature)
        self.assertEqual((self.neg*self.unknown_curv).curvature, QUASILINEAR)
        self.assertEqual((self.pos*self.aff).curvature, self.aff.curvature)
        self.assertEqual((self.pos*self.ccv).curvature, self.ccv.curvature)
        self.assertEqual((self.unknown_sign*self.const).curvature, self.const.curvature)
        self.assertEqual((self.unknown_sign*self.ccv).curvature, UNKNOWN)

    def test_neg(self) -> None:
        self.assertEqual((-self.cvx).curvature, self.ccv.curvature)
        self.assertEqual((-self.aff).curvature, self.aff.curvature)

    # Tests the is_affine, is_convex, and is_concave methods
    def test_is_curvature(self) -> None:
        assert self.const.is_affine()
        assert self.aff.is_affine()
        assert not self.cvx.is_affine()
        assert not self.ccv.is_affine()
        assert not self.unknown_curv.is_affine()

        assert self.const.is_convex()
        assert self.aff.is_convex()
        assert self.cvx.is_convex()
        assert not self.ccv.is_convex()
        assert not self.unknown_curv.is_convex()

        assert self.const.is_concave()
        assert self.aff.is_concave()
        assert not self.cvx.is_concave()
        assert self.ccv.is_concave()
        assert not self.unknown_curv.is_concave()



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

from typing import Tuple

import numpy as np

from cvxpy.atoms.atom import Atom
from cvxpy.expressions.constants.parameter import is_param_free
from cvxpy.expressions.expression import Expression
from cvxpy.expressions.variable import Variable
from cvxpy.utilities import scopes


class perspective(Atom):
    r"""Implements the perspective transform of a convex or concave scalar
    expression. Uses the fact that given a cone form for the epigraph of :math:`f` via

    :math:`\{ (t, x) \in \mathbb{R}^{n+1} : t \geq f(x) \}`
    :math:`= \{ (t,x) : Fx + gt + e \in K \},`

    the epigraph of the perspective transform
    of f can be given by

    :math:`\{ (t, x, s) \in \mathbb{R}^{n+2} : t \geq sf(x/s) \}`
    :math:`= \{ (t,x,s) : Fx + gt + se \in K \},`

    (see https://web.stanford.edu/~boyd/papers/pdf/sw_aff_ctrl.pdf).

    Note that this is the perspective transform of a scalar expression viewed as
    a function of its underlying variables. The perspective atom does not return
    a `Callable`, so you cannot create compositions such as :math:`p(g(x),s)`, where
    :math:`p(z,s) = sf(z/s)` is the perpective transform of :math:`f`.
    """

    def __init__(self, f: Expression, s: Variable, f_recession: Expression = None) -> None:
        self.f = f
        self.f_recession = f_recession
        super(perspective, self).__init__(s, *f.variables())

    def validate_arguments(self) -> None:
        assert self.f.size == 1  # dealing only with scalars, for now
        assert self.args[0].size == 1
        assert isinstance(self.args[0], Variable), "s must be a variable"
        assert self.args[0].is_nonneg(), "s must be a nonnegative variable"
        return super().validate_arguments()

    def numeric(self, values: list[np.ndarray, np.ndarray]) -> np.ndarray:
        """
        Compute the perspective sf(x/s) numerically.
        """

        assert values[0] >= 0
        s_val = np.array(values[0])

        # Normal scenario: compute the perspective transform of the given function f
        f = self.f
        if np.isclose(values[0], 0.0):
            # Handle s = 0 with the recession function by swapping f with f_recession
            # Since we just swap the two functions, we end up with s * f_recession(x / s) 
            # when we actually just want f_recession(x). Thus we set s=1 to ignore s.
            assert self.f_recession is not None, (
                "To handle s = 0, pass in a recession function f_recession"
            )
            f = self.f_recession
            values[0] = 1 

        old_x_vals = [var.value for var in f.variables()]

        def set_vals(vals, s_val):
            for var, val in zip(f.variables(), vals):
                var.value = val/s_val

        set_vals(values[1:], s_val=values[0])

        ret_val = np.array([f.value*s_val])

        set_vals(old_x_vals, s_val=1)

        return ret_val

    def sign_from_args(self) -> tuple[bool, bool]:
        f_pos = self.f.is_nonneg()
        f_neg = self.f.is_nonpos()
        s_pos = self.args[0].is_nonneg()

        assert s_pos

        is_positive = (f_pos and s_pos)
        is_negative = (f_neg and s_pos)

        return is_positive, is_negative

    def is_atom_convex(self) -> bool:
        """Is the atom convex?
        """
        if scopes.dpp_scope_active() and not is_param_free(self.f):
            return False
        else:
            return self.f.is_convex() and self.args[0].is_nonneg()

    def is_atom_concave(self) -> bool:
        """Is the atom concave?
        """
        if scopes.dpp_scope_active() and not is_param_free(self.f):
            return False
        else:
            return self.f.is_concave() and self.args[0].is_nonneg()

    def is_incr(self, idx) -> bool:
        """Is the composition non-decreasing in argument idx?
        """
        return False

    def is_decr(self, idx: int) -> bool:
        """Is the composition non-increasing in argument idx?
        """
        return False

    def shape_from_args(self) -> Tuple[int, ...]:
        """Returns the (row, col) shape of the expression.
        """
        return self.f.shape

    def _grad(self, values):
        """Gives the (sub/super)gradient of the atom w.r.t. each argument.

        Matrix expressions are vectorized, so the gradient is a matrix.

        Args:
            values: A list of numeric values for the arguments.

        Returns:
            A list of SciPy CSC sparse matrices or None.
        """
        # TODO
        raise NotImplementedError()



"""
Copyright 2025, the CVXPY Authors

Licensed under the Apache License, Version 2.0 (the "License");

you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

CITATION_DICT = {}

CITATION_DICT["CVXPY"] = \
"""
@article{diamond2016cvxpy,
  author  = {Steven Diamond and Stephen Boyd},
  title   = {{CVXPY}: {A} {P}ython-embedded modeling language for convex optimization},
  journal = {Journal of Machine Learning Research},
  year    = {2016},
  volume  = {17},
  number  = {83},
  pages   = {1--5}
}

@article{agrawal2018rewriting,
  author  = {Agrawal, Akshay and Verschueren, Robin and Diamond, Steven and Boyd, Stephen},
  title   = {A rewriting system for convex optimization problems},
  journal = {Journal of Control and Decision},
  year    = {2018},
  volume  = {5},
  number  = {1},
  pages   = {42--60}
}
"""

CITATION_DICT["DCP"] = \
"""
@book{grant2006disciplined,
  title={Disciplined convex programming},
  author={Grant, Michael and Boyd, Stephen and Ye, Yinyu},
  year={2006},
  publisher={Springer}
}
"""

CITATION_DICT["DGP"] = \
"""
@article{agrawal2019disciplined,
  title={Disciplined geometric programming},
  author={Agrawal, Akshay and Diamond, Steven and Boyd, Stephen},
  journal={Optimization Letters},
  volume={13},
  pages={961--976},
  year={2019},
  publisher={Springer}
}
"""

CITATION_DICT["DQCP"] = \
"""
@article{agrawal2020disciplined,
  title={Disciplined quasiconvex programming},
  author={Agrawal, Akshay and Boyd, Stephen},
  journal={Optimization Letters},
  volume={14},
  number={7},
  pages={1643--1657},
  year={2020},
  publisher={Springer}
}
"""

CITATION_DICT["CVXOPT"] = \
"""
@article{vandenberghe2010cvxopt,
  title={The CVXOPT linear and quadratic cone program solvers},
  author={Vandenberghe, Lieven},
  journal={Online: http://cvxopt. org/documentation/coneprog. pdf},
  volume={53},
  year={2010}
}
"""

CITATION_DICT["GLPK"] = \
"""
@inproceedings{Oki2012GLPKL,
  title={GLPK (GNU Linear Programming Kit)},
  author={Eiji Oki},
  year={2012},
  url={https://api.semanticscholar.org/CorpusID:63578694}
}
"""

CITATION_DICT["GLPK"] = \
"""
@inproceedings{Oki2012GLPKL,
  title={GLPK (GNU Linear Programming Kit)},
  author={Eiji Oki},
  year={2012},
  url={https://api.semanticscholar.org/CorpusID:63578694}
}
"""

CITATION_DICT["GLPK_MI"] = \
"""
@inproceedings{Oki2012GLPKL,
  title={GLPK (GNU Linear Programming Kit)},
  author={Eiji Oki},
  year={2012},
  url={https://api.semanticscholar.org/CorpusID:63578694}
}
"""

CITATION_DICT["GLOP"] = \
"""
@software{ortools,
  title = {OR-Tools},
  version = { v9.11 },
  author = {Laurent Perron and Vincent Furnon},
  organization = {Google},
  url = {https://developers.google.com/optimization/},
  date = { 2024-05-07 }
}
"""

CITATION_DICT["CBC"] = \
"""
@inbook{Forrest2005,
  title = {CBC User Guide},
  ISBN = {9781877640216},
  url = {http://dx.doi.org/10.1287/educ.1053.0020},
  DOI = {10.1287/educ.1053.0020},
  booktitle = {Emerging Theory,  Methods,  and Applications},
  publisher = {INFORMS},
  author = {Forrest,  John and Lougee-Heimer,  Robin},
  year = {2005},
  month = sep,
  pages = {257–277}
}
"""

CITATION_DICT["COPT"] = \
"""
@article{ge2022cardinal,
  title={Cardinal Optimizer (COPT) user guide},
  author={Ge, Dongdong and Huangfu, Qi and Wang, Zizhuo and Wu, Jian and Ye, Yinyu},
  journal={arXiv preprint arXiv:2208.14314},
  year={2022}
}
"""

CITATION_DICT["ECOS"] = \
"""
@inproceedings{domahidi2013ecos,
  title={ECOS: An SOCP solver for embedded systems},
  author={Domahidi, Alexander and Chu, Eric and Boyd, Stephen},
  booktitle={2013 European control conference (ECC)},
  pages={3071--3076},
  year={2013},
  organization={IEEE}
}
"""

CITATION_DICT["ECOS_EXP"] = \
"""
@book{serrano2015algorithms,
  title={Algorithms for unsymmetric cone optimization and an implementation for problems
  with the exponential cone},
  author={Serrano, Santiago Akle},
  year={2015},
  publisher={Stanford University}
}
"""

CITATION_DICT["SCS"] = \
"""
@article{odonoghue2021operator,
  title={Operator splitting for a homogeneous embedding of the linear complementarity problem},
  author={O'Donoghue, Brendan},
  journal={SIAM Journal on Optimization},
  volume={31},
  number={3},
  pages={1999--2023},
  year={2021},
  publisher={SIAM}
}
"""

CITATION_DICT["SDPA"] = \
"""
@article{doi:10.1080/1055678031000118482,
    author    = {Yamashita, Makoto
                 and Fujisawa, Katsuki
                 and Kojima, Masakazu},
    title     = {Implementation and evaluation of SDPA 6.0 
    (Semidefinite Programming Algorithm 6.0)},
    journal   = {Optimization Methods and Software},
    volume    = {18},
    number    = {4},
    pages     = {491-505},
    year      = {2003},
    publisher = {Taylor & Francis},
    doi       = {10.1080/1055678031000118482},
    URL       = {https://doi.org/10.1080/1055678031000118482},
    eprint    = {https://doi.org/10.1080/1055678031000118482}
}

@Inbook{Yamashita2012,
    author    = {Yamashita, Makoto
                 and Fujisawa, Katsuki
                 and Fukuda, Mituhiro
                 and Kobayashi, Kazuhiro
                 and Nakata, Kazuhide
                 and Nakata, Maho},
    editor    = {Anjos, Miguel F.
                 and Lasserre, Jean B.},
    title           = {Latest Developments in the SDPA Family for Solving Large-Scale SDPs},
    bookTitle       = {Handbook on Semidefinite, Conic and Polynomial Optimization},
    year      = {2012},
    publisher = {Springer US},
    address   = {Boston, MA},
    pages     = {687--713},
    isbn      = {978-1-4614-0769-0},
    doi       = {10.1007/978-1-4614-0769-0_24},
    url       = {https://doi.org/10.1007/978-1-4614-0769-0_24}
}

@inproceedings{doi:10.1109/CACSD.2010.5612693,
    author    = {Nakata, Maho},
    booktitle = {2010 IEEE International Symposium on Computer-Aided Control System Design}, 
    title     = {A numerical evaluation of highly accurate multiple-precision arithmetic version
    of semidefinite programming solver: SDPA-GMP, -QD and -DD.}, 
    year      = {2010},
    volume    = {},
    number    = {},
    pages     = {29-34},
    doi       = {10.1109/CACSD.2010.5612693}
}

@article{Kim2011,
    author    = {Kim, Sunyoung
                 and Kojima, Masakazu
                 and Mevissen, Martin
                 and Yamashita, Makoto},
    title     = {Exploiting sparsity in linear and nonlinear matrix inequalities
    via positive semidefinite matrix completion},
    journal   = {Mathematical Programming},
    year      = {2011},
    month     = {Sep},
    day       = {01},
    volume    = {129},
    number    = {1},
    pages     = {33-68},
    issn      = {1436-4646},
    doi       = {10.1007/s10107-010-0402-6},
    url       = {https://doi.org/10.1007/s10107-010-0402-6}
}
"""

CITATION_DICT["DIFFCP"] = \
"""
@misc{agrawal2020differentiating,
    title={Differentiating Through a Cone Program}, 
    author={Akshay Agrawal and Shane Barratt and Stephen Boyd and Enzo Busseti and Walaa M. Moursi},
    year={2020},
    eprint={1904.09043},
    archivePrefix={arXiv},
    primaryClass={math.OC},
    url={https://arxiv.org/abs/1904.09043}, 
}
"""

CITATION_DICT["GUROBI"] = \
"""
@manual{gurobi,
  author = {Gurobi Optimization, LLC},
  title  = {Gurobi Optimizer Reference Manual},
  year   = 2025,
  url    = {https://www.gurobi.com}
}
"""

CITATION_DICT["OSQP"] = \
"""
@article{osqp,
  author  = {Stellato, B. and Banjac, G. and Goulart, P. and Bemporad, A. and Boyd, S.},
  title   = {{OSQP}: an operator splitting solver for quadratic programs},
  journal = {Mathematical Programming Computation},
  year    = {2020},
  volume  = {12},
  number  = {4},
  pages   = {637--672},
  doi     = {10.1007/s12532-020-00179-2},
  url     = {https://doi.org/10.1007/s12532-020-00179-2}
}
"""

CITATION_DICT["PIQP"] = \
"""
@inproceedings{schwan2023piqp,
  author    = {Schwan, Roland and Jiang, Yuning and Kuhn, Daniel and Jones, Colin N.},
  booktitle = {2023 62nd IEEE Conference on Decision and Control (CDC)},
  title     = {{PIQP}: A Proximal Interior-Point Quadratic Programming Solver},
  year      = {2023},
  volume    = {},
  number    = {},
  pages     = {1088-1093},
  doi       = {10.1109/CDC49753.2023.10383915}
}
"""

CITATION_DICT["PROXQP"] = \
"""
@inproceedings{bambade2022prox,
  title={Prox-qp: Yet another quadratic programming solver for robotics and beyond},
  author={Bambade, Antoine and El-Kazdadi, Sarah and Taylor, Adrien and Carpentier, Justin},
  booktitle={RSS 2022-Robotics: Science and Systems},
  year={2022}
}
"""

CITATION_DICT["QPALM"] = \
"""
@article{hermans2022qpalm,
	author = {Hermans, Ben and Themelis, Andreas and Patrinos, Panagiotis},
	title = {{QPALM}: a proximal augmented lagrangian method for nonconvex quadratic programs},
	volume = {14},
	issn = {1867-2957},
	pages = {497--541},
	number = {3},
	journal = {Mathematical Programming Computation},
	month = sep,
	year = {2022},
	doi = {10.1007/s12532-022-00218-0},
}
"""

CITATION_DICT["QOCO"] = \
r"""
@misc{chari2025qoco,
  title         = {QOCO: A Quadratic Objective Conic Optimizer with Custom Solver Generation},
  author        = {Chari, Govind M and A{\c{c}}{\i}kme{\c{s}}e, Beh{\c{c}}et},
  year          = {2025},
  eprint        = {2503.12658},
  archiveprefix = {arXiv},
  primaryclass  = {math.OC},
  url           = {https://arxiv.org/abs/2503.12658}
}
"""

CITATION_DICT["CPLEX"] = \
"""
@article{manual1987ibm,
  title={Ibm ilog cplex optimization studio},
  author={Manual, CPLEX User’s},
  journal={Version},
  volume={12},
  number={1987-2018},
  pages={1},
  year={1987}
}
"""

CITATION_DICT["MOSEK"] = \
"""
@manual{mosek,
   author = {MOSEK ApS},
   title = {MOSEK Optimization Suite},
   year = {2025},
   url = {https://docs.mosek.com/latest/intro/index.html}
 }
"""

CITATION_DICT["MOSEK_EXP"] = \
"""
@article{Dahl2021,
  title = {A primal-dual interior-point algorithm for nonsymmetric exponential-cone optimization},
  volume = {194},
  ISSN = {1436-4646},
  url = {http://dx.doi.org/10.1007/s10107-021-01631-4},
  DOI = {10.1007/s10107-021-01631-4},
  number = {1–2},
  journal = {Mathematical Programming},
  publisher = {Springer Science and Business Media LLC},
  author = {Dahl,  Joachim and Andersen,  Erling D.},
  year = {2021},
  month = mar,
  pages = {341–370}
}
"""

CITATION_DICT["XPRESS"] = \
"""
@manual{xpress,
  author = {FICO},
  title  = {FICO Xpress Optimization},
  year   = {2025},
  url    = {https://www.fico.com/fico-xpress-optimization/docs/latest/overview.html}
}
"""

CITATION_DICT["NAG"] = \
"""
@misc{xpress,
  author = {NAG},
  title  = {Optimization Modelling Suite},
  year   = {2025},
  url    = {https://nag.com/mathematical-optimization/}
}
"""

CITATION_DICT["PDLP"] = \
"""
@article{applegate2021practical,
  title={Practical large-scale linear programming using primal-dual hybrid gradient},
  author={Applegate, David and D{\'\\i}az, Mateo and Hinder, Oliver and Lu, Haihao and Lubin, Miles
  and O'Donoghue, Brendan and Schudy, Warren},
  journal={Advances in Neural Information Processing Systems},
  volume={34},
  pages={20243--20257},
  year={2021}
}

@article{applegate2025pdlp,
  title={PDLP: A Practical First-Order Method for Large-Scale Linear Programming},
  author={Applegate, David and D{\'\\i}az, Mateo and Hinder, Oliver and Lu, Haihao and Lubin, Miles
  and O'Donoghue, Brendan and Schudy, Warren},
  journal={arXiv preprint arXiv:2501.07018},
  year={2025}
}
"""

CITATION_DICT["SCIP"] = \
"""
@article{bolusani2024scip,
  title={The SCIP optimization suite 9.0},
  author={Bolusani, Suresh and Besan{\\c{c}}on, Mathieu and Bestuzheva, Ksenia
  and Chmiela, Antonia and Dion{\'\\i}sio, Jo{\\~a}o and Donkiewicz, Tim
  and van Doornmalen, Jasper and Eifler, Leon and Ghannam, Mohammed and Gleixner, Ambros
  and others},
  journal={arXiv preprint arXiv:2402.17702},
  year={2024}
}
"""

CITATION_DICT["SCIPY"] = \
"""
@ARTICLE{2020SciPy-NMeth,
  author  = {Virtanen, Pauli and Gommers, Ralf and Oliphant, Travis E. and
            Haberland, Matt and Reddy, Tyler and Cournapeau, David and
            Burovski, Evgeni and Peterson, Pearu and Weckesser, Warren and
            Bright, Jonathan and {van der Walt}, St{\'e}fan J. and
            Brett, Matthew and Wilson, Joshua and Millman, K. Jarrod and
            Mayorov, Nikolay and Nelson, Andrew R. J. and Jones, Eric and
            Kern, Robert and Larson, Eric and Carey, C J and
            Polat, {\\.I}lhan and Feng, Yu and Moore, Eric W. and
            {VanderPlas}, Jake and Laxalde, Denis and Perktold, Josef and
            Cimrman, Robert and Henriksen, Ian and Quintero, E. A. and
            Harris, Charles R. and Archibald, Anne M. and
            Ribeiro, Ant{\\^o}nio H. and Pedregosa, Fabian and
            {van Mulbregt}, Paul and {SciPy 1.0 Contributors}},
  title   = {{{SciPy} 1.0: Fundamental Algorithms for Scientific
            Computing in Python}},
  journal = {Nature Methods},
  year    = {2020},
  volume  = {17},
  pages   = {261--272},
  adsurl  = {https://rdcu.be/b08Wh},
  doi     = {10.1038/s41592-019-0686-2},
}
"""

CITATION_DICT["CLARABEL"] = \
"""
@misc{Clarabel_2024,
      title={Clarabel: An interior-point solver for conic programs with quadratic objectives}, 
      author={Paul J. Goulart and Yuwen Chen},
      year={2024},
      eprint={2405.12762},
      archivePrefix={arXiv},
      primaryClass={math.OC}
}
"""

CITATION_DICT["CUCLARABEL"] = \
"""
@misc{CuClarabel,
      title={CuClarabel: GPU Acceleration for a Conic Optimization Solver}, 
      author={Yuwen Chen and Danny Tse and Parth Nobel and Paul Goulart and Stephen Boyd},
      year={2024},
      eprint={2412.19027},
      archivePrefix={arXiv},
      primaryClass={math.OC},
      url={https://arxiv.org/abs/2412.19027}, 
}
"""

CITATION_DICT["DAQP"] = \
"""
@article{arnstrom2022dual,
  title={A Dual Active-Set Solver for Embedded Quadratic Programming Using Recursive
  LDL $\\^{}$\\{$T$\\}$ $ Updates},
  author={Arnstr{\"o}m, Daniel and Bemporad, Alberto and Axehill, Daniel},
  journal={IEEE Transactions on Automatic Control},
  volume={67},
  number={8},
  pages={4362--4369},
  year={2022},
  publisher={IEEE}
}
"""

CITATION_DICT["HIGHS"] = \
"""
@article{Huangfu2017,
  title = {Parallelizing the dual revised simplex method},
  volume = {10},
  ISSN = {1867-2957},
  url = {http://dx.doi.org/10.1007/s12532-017-0130-5},
  DOI = {10.1007/s12532-017-0130-5},
  number = {1},
  journal = {Mathematical Programming Computation},
  publisher = {Springer Science and Business Media LLC},
  author = {Huangfu,  Q. and Hall,  J. A. J.},
  year = {2017},
  month = dec,
  pages = {119–142}
}
"""

CITATION_DICT["COSMO"] = \
"""
@Article{Garstka_2021,
  author  = {Michael Garstka and Mark Cannon and Paul Goulart},
  journal = {Journal of Optimization Theory and Applications},
  title   = {{COSMO}: A Conic Operator Splitting Method for Convex Conic Problems},
  volume  = {190},
  number  = {3},
  pages   = {779--810},
  year    = {2021},
  publisher = {Springer},
  doi     = {10.1007/s10957-021-01896-x},
  url     = {https://doi.org/10.1007/s10957-021-01896-x}
}
"""

CITATION_DICT["MPAX"] = \
"""
@article{lu2024mpax,
  title={MPAX: Mathematical Programming in JAX},
  author={Lu, Haihao and Peng, Zedong and Yang, Jinwen},
  journal={arXiv preprint arXiv:2412.09734},
  year={2024}
}
"""

CITATION_DICT["KNITRO"] = \
"""
@article{byrd2006k,
  title={K nitro: An integrated package for nonlinear optimization},
  author={Byrd, Richard H and Nocedal, Jorge and Waltz, Richard A},
  journal={Large-scale nonlinear optimization},
  pages={35--59},
  year={2006},
  publisher={Springer}
}
"""

CITATION_DICT["CUOPT"] = \
"""
@software{cuOpt,
  title = {cuOpt},
  version = { 25.05 },
  organization = {NVIDIA},
  url = {https://docs.nvidia.com/cuopt/index.html},
  date = { 2025-05-29 }
}
"""

CITATION_DICT["MOREAU"] = \
"""
@misc{moreau2025,
  title={Moreau: GPU-Accelerated Conic Optimization},
  year={2025},
  note={https://www.optimalintellect.com/}
}
"""



"""
Copyright 2025 CVXPY developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import pytest

import cvxpy as cp
from cvxpy.constraints.power import PowConeND
from cvxpy.reductions.solvers.conic_solvers.scs_conif import SCS as SCS_SOLVER


def solve_prob(prob, solver):
    prob.solve(solver=solver, verbose=False)
    assert prob.status in [cp.OPTIMAL, cp.OPTIMAL_INACCURATE]


class TestPowConeND:
    """Unit tests for PowConeND and PowCone3D."""

    @pytest.mark.parametrize("axis", [0, 1])
    @pytest.mark.parametrize("solver", [cp.CLARABEL, cp.SCS])
    def test_pow_cone_nd_3d(self, axis, solver) -> None:
        """
        A variation of test_pcp_3.
        Clarabel natively supports ND power cones; SCS only supports 3D power
        cones so CVXPY applies a reduction.
        """
        # SCS does not natively support PowConeND, so the decomposition
        # to PowCone3D is exercised when solving with SCS.
        assert PowConeND not in SCS_SOLVER.SUPPORTED_CONSTRAINTS
        expect_x = np.array([0.06393515, 0.78320961, 2.30571048])
        x = cp.Variable(3, name='x')
        hypos = cp.Variable(2, name='hypos')
        objective = cp.Maximize(cp.sum(hypos) - x[0])
        W = cp.bmat([[x[0], x[2]],
                     [x[1], 1.0]])
        alpha = np.array([[0.2, 0.4],
                          [0.8, 0.6]])
        if axis == 1:
            W = W.T
            alpha = alpha.T

        constraints = [
            x[0] + x[1] + 0.5 * x[2] == 2,
            cp.constraints.PowConeND(W, hypos, alpha, axis=axis)
        ]
        prob = cp.Problem(objective, constraints)
        solve_prob(prob, solver)
        np.testing.assert_allclose(x.value, expect_x, atol=1e-3)

    @pytest.mark.parametrize("axis", [0, 1])
    @pytest.mark.parametrize("solver", [cp.CLARABEL, cp.SCS])
    def test_pow_cone_nd_3d_variable_swap(self, axis, solver) -> None:
        """
        A variation of test_pcp_3 with variable appearing in a different order.
        """
        expect_x = np.array([0.06393515, 2.30571048, 0.78320961])
        x = cp.Variable(3)
        hypos = cp.Variable(2)
        objective = cp.Maximize(cp.sum(hypos) - x[0])
        W = cp.bmat([[x[0], x[1]],
                     [x[2], 1.0]])
        alpha = np.array([[0.2, 0.4],
                          [0.8, 0.6]])
        if axis == 1:
            W = W.T
            alpha = alpha.T
        constraints = [
            x[0] + x[2] + 0.5 * x[1] == 2,
            cp.constraints.PowConeND(W, hypos, alpha, axis=axis)
        ]
        prob = cp.Problem(objective, constraints)
        solve_prob(prob, solver)
        np.testing.assert_allclose(x.value, expect_x, atol=1e-3)

    @pytest.mark.parametrize("axis", [0, 1])
    @pytest.mark.parametrize("solver", [cp.CLARABEL, cp.SCS])
    def test_pow_cone_nd(self, axis, solver) -> None:
        """Solving a PowConeND constraint with more than 3 dimensions."""
        expect_x = np.array([0, 0, 0, 2.28571379, 3.42857186])
        x = cp.Variable(5)
        hypos = cp.Variable(2)
        objective = cp.Maximize(cp.sum(hypos) - x[0])
        W = cp.bmat([[x[0], x[3]],
                     [x[1], x[4]],
                     [x[2], 1.0]])
        alpha = np.array([[0.2, 0.4],
                          [0.4, 0.3],
                          [0.4, 0.3]])
        if axis == 1:
            W = W.T
            alpha = alpha.T
        constraints = [
            x[0] + x[1] + 0.5 * x[2] + 0.5 * x[3] + 0.25 * x[4] == 2,
            cp.constraints.PowConeND(W, hypos, alpha, axis=axis)
        ]
        prob = cp.Problem(objective, constraints)
        solve_prob(prob, solver)
        np.testing.assert_allclose(x.value, expect_x, atol=1e-3)

    @pytest.mark.parametrize("axis", [0, 1])
    @pytest.mark.parametrize("solver", [cp.CLARABEL, cp.SCS])
    def test_pow_cone_nd_variable_swap(self, axis, solver) -> None:
        """
        A variation of test_pow_cone_nd with variable appearing in a different order.
        """
        expect_x = np.array([3.42857186, 0, 0, 2.28571379, 0])
        x = cp.Variable(5)
        hypos = cp.Variable(2)
        objective = cp.Maximize(cp.sum(hypos) - x[4])
        W = cp.bmat([[x[4], x[3]],
                     [x[1], x[0]],
                     [x[2], 1.0]])
        alpha = np.array([[0.2, 0.4],
                          [0.4, 0.3],
                          [0.4, 0.3]])
        if axis == 1:
            W = W.T
            alpha = alpha.T
        constraints = [
            x[4] + x[1] + 0.5 * x[2] + 0.5 * x[3] + 0.25 * x[0] == 2,
            cp.constraints.PowConeND(W, hypos, alpha, axis=axis)
        ]
        prob = cp.Problem(objective, constraints)
        solve_prob(prob, solver)
        np.testing.assert_allclose(x.value, expect_x, atol=1e-3)

    @pytest.mark.parametrize("axis", [0, 1])
    @pytest.mark.parametrize("solver", [cp.CLARABEL, cp.SCS])
    def test_pow_cone_nd_single_cone(self, axis, solver) -> None:
        """
        Solving a PowConeND constraint with only a single cone.
        Ensures no variables collapse to lower dimensions incorrectly.
        """
        x = cp.Variable(2)
        hypos = cp.Variable(1)
        objective = cp.Maximize(cp.sum(hypos) - x[0])
        W = cp.bmat([[x[0]], [x[1]]])
        alpha = np.array([[0.2], [0.8]])
        if axis == 1:
            W = W.T
            alpha = alpha.T
        constraints = [
            x[0] + x[1] == 2,
            cp.constraints.PowConeND(W, hypos, alpha, axis=axis)
        ]
        prob = cp.Problem(objective, constraints)
        solve_prob(prob, solver)

    @pytest.mark.parametrize("solver", [cp.CLARABEL, cp.SCS])
    def test_3d_pow_cone_scalar_alpha(self, solver) -> None:
        """Test PowCone3D with scalar alpha."""
        x = cp.Variable(3)
        constraints = [cp.PowCone3D(x[0], x[1], x[2], 0.75)]
        prob = cp.Problem(cp.Minimize(cp.norm(x)), constraints)
        solve_prob(prob, solver)

    @pytest.mark.parametrize("n", [3, 4, 5, 6, 7, 8, 9, 10, 16, 32, 100])
    @pytest.mark.parametrize("solver,atol", [(cp.CLARABEL, 1e-4), (cp.SCS, 1e-3)])
    def test_pow_cone_nd_balanced_tree_decomposition(self, n, solver, atol) -> None:
        """Test balanced tree decomposition for various n values."""
        W = cp.Variable(n, pos=True)
        z = cp.Variable()
        alpha = np.ones(n) / n
        con = cp.constraints.PowConeND(W, z, alpha)
        prob = cp.Problem(cp.Maximize(z), [con, W <= 2])
        solve_prob(prob, solver)
        np.testing.assert_allclose(z.value, 2.0, atol=atol)

    @pytest.mark.parametrize("solver,atol", [(cp.CLARABEL, 1e-4), (cp.SCS, 1e-3)])
    def test_pow_cone_nd_balanced_tree_nonuniform_alpha(self, solver, atol) -> None:
        """Test balanced tree decomposition with non-uniform alpha weights."""
        alpha = np.array([0.05, 0.1, 0.15, 0.2, 0.2, 0.15, 0.1, 0.05])
        bounds = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])
        expected = np.prod(bounds ** alpha)

        W = cp.Variable(8, pos=True)
        z = cp.Variable()
        con = cp.constraints.PowConeND(W, z, alpha)
        prob = cp.Problem(cp.Maximize(z), [con, W <= bounds])
        solve_prob(prob, solver)
        np.testing.assert_allclose(z.value, expected, atol=atol)

    @pytest.mark.parametrize("solver,atol", [(cp.CLARABEL, 1e-4), (cp.SCS, 1e-3)])
    def test_pow_cone_nd_balanced_tree_multiple_cones(self, solver, atol) -> None:
        """Test balanced tree decomposition with multiple cones (k > 1)."""
        n, k = 8, 3
        W = cp.Variable((n, k), pos=True)
        z = cp.Variable(k)
        alpha = np.ones((n, k)) / n
        con = cp.constraints.PowConeND(W, z, alpha, axis=0)
        prob = cp.Problem(cp.Maximize(cp.sum(z)), [con, W <= 2])
        solve_prob(prob, solver)
        np.testing.assert_allclose(z.value, [2.0, 2.0, 2.0], atol=atol)

    @pytest.mark.parametrize("solver,atol", [(cp.CLARABEL, 1e-4), (cp.SCS, 1e-3)])
    def test_pow_cone_nd_balanced_tree_multiple_cones_nonuniform(self, solver, atol) -> None:
        """Test balanced tree with non-uniform alpha and multiple cones (k > 1)."""
        n, k = 6, 2
        bounds = np.array([
            [1.0, 2.0],
            [2.0, 3.0],
            [3.0, 4.0],
            [4.0, 5.0],
            [5.0, 6.0],
            [6.0, 7.0],
        ])
        alpha = np.array([
            [0.1, 0.3],
            [0.2, 0.2],
            [0.3, 0.1],
            [0.15, 0.15],
            [0.15, 0.1],
            [0.1, 0.15],
        ])
        expected = np.prod(bounds ** alpha, axis=0)

        W = cp.Variable((n, k), pos=True)
        z = cp.Variable(k)
        con = cp.constraints.PowConeND(W, z, alpha, axis=0)
        prob = cp.Problem(cp.Maximize(cp.sum(z)), [con, W <= bounds])
        solve_prob(prob, solver)
        np.testing.assert_allclose(z.value, expected, atol=atol)

    @pytest.mark.parametrize("n,rtol", [
        (2, 1e-3),   # n=2: no tree needed
        (3, 1e-2),   # n=3: depth 1
        (4, 1e-2),   # n=4: depth 2
        (8, 1e-2),   # n=8: depth 3
        (16, 5e-2),  # n=16: depth 4
    ])
    def test_pow_cone_nd_dual_variables(self, n, rtol) -> None:
        """Test dual variable recovery matches CLARABEL."""
        W = cp.Variable((n, 1), pos=True)
        z = cp.Variable(1)
        alpha = np.ones((n, 1)) / n
        con = cp.constraints.PowConeND(W, z, alpha, axis=0)
        prob = cp.Problem(cp.Maximize(z[0]), [con, W <= 2])

        solve_prob(prob, cp.CLARABEL)
        clarabel_w_dual = con.dual_value[0].flatten()
        clarabel_z_dual = con.dual_value[1].flatten()

        solve_prob(prob, cp.SCS)
        scs_w_dual = con.dual_value[0].flatten()
        scs_z_dual = con.dual_value[1].flatten()

        np.testing.assert_allclose(scs_w_dual, clarabel_w_dual, rtol=rtol)
        np.testing.assert_allclose(scs_z_dual, clarabel_z_dual, rtol=rtol)



import numpy as np

from cvxpy.atoms.affine.add_expr import AddExpression
from cvxpy.atoms.affine.binary_operators import (DivExpression, MulExpression,
                                                 multiply,)
from cvxpy.atoms.affine.sum import Sum
from cvxpy.atoms.affine.trace import Trace
from cvxpy.atoms.cumprod import cumprod
from cvxpy.atoms.elementwise.exp import exp
from cvxpy.atoms.elementwise.log import log
from cvxpy.atoms.elementwise.maximum import maximum
from cvxpy.atoms.elementwise.minimum import minimum
from cvxpy.atoms.elementwise.power import Power, PowerApprox
from cvxpy.atoms.elementwise.xexp import xexp
from cvxpy.atoms.eye_minus_inv import eye_minus_inv
from cvxpy.atoms.geo_mean import GeoMean, GeoMeanApprox
from cvxpy.atoms.gmatmul import gmatmul
from cvxpy.atoms.max import max
from cvxpy.atoms.min import min
from cvxpy.atoms.norm1 import norm1
from cvxpy.atoms.norm_inf import norm_inf
from cvxpy.atoms.one_minus_pos import one_minus_pos
from cvxpy.atoms.pf_eigenvalue import pf_eigenvalue
from cvxpy.atoms.pnorm import Pnorm, PnormApprox
from cvxpy.atoms.prod import Prod
from cvxpy.atoms.quad_form import quad_form
from cvxpy.atoms.quad_over_lin import quad_over_lin
from cvxpy.constraints.finite_set import FiniteSet
from cvxpy.expressions.constants.constant import Constant
from cvxpy.expressions.constants.parameter import Parameter
from cvxpy.expressions.variable import Variable
from cvxpy.reductions.dgp2dcp.canonicalizers.add_canon import add_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.constant_canon import (
    constant_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.cumprod_canon import (
    cumprod_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.div_canon import div_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.exp_canon import exp_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.eye_minus_inv_canon import (
    eye_minus_inv_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.finite_set_canon import (
    finite_set_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.geo_mean_canon import (
    geo_mean_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.gmatmul_canon import gmatmul_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.log_canon import log_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.mul_canon import mul_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.mulexpression_canon import (
    mulexpression_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.norm1_canon import norm1_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.norm_inf_canon import (
    norm_inf_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.one_minus_pos_canon import (
    one_minus_pos_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.pf_eigenvalue_canon import (
    pf_eigenvalue_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.pnorm_canon import pnorm_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.power_canon import power_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.prod_canon import prod_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.quad_form_canon import (
    quad_form_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.quad_over_lin_canon import (
    quad_over_lin_canon,)
from cvxpy.reductions.dgp2dcp.canonicalizers.sum_canon import sum_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.trace_canon import trace_canon
from cvxpy.reductions.dgp2dcp.canonicalizers.xexp_canon import xexp_canon
from cvxpy.reductions.eliminate_pwl.canonicalizers import (
    CANON_METHODS as PWL_METHODS,)

CANON_METHODS = {
    AddExpression : add_canon,
    Constant : constant_canon,
    cumprod : cumprod_canon,
    DivExpression : div_canon,
    exp : exp_canon,
    eye_minus_inv : eye_minus_inv_canon,
    FiniteSet : finite_set_canon,
    GeoMean : geo_mean_canon,
    GeoMeanApprox : geo_mean_canon,
    gmatmul : gmatmul_canon,
    log : log_canon,
    MulExpression : mulexpression_canon,
    multiply : mul_canon,
    norm1 : norm1_canon,
    norm_inf : norm_inf_canon,
    one_minus_pos : one_minus_pos_canon,
    pf_eigenvalue : pf_eigenvalue_canon,
    Pnorm : pnorm_canon,
    PnormApprox : pnorm_canon,
    Power : power_canon,
    PowerApprox : power_canon,
    Prod : prod_canon,
    quad_form : quad_form_canon,
    quad_over_lin : quad_over_lin_canon,
    Trace : trace_canon,
    Sum : sum_canon,
    xexp : xexp_canon,
    Variable : None,
    Parameter : None,
}

CANON_METHODS[max] = PWL_METHODS[max]
CANON_METHODS[min] = PWL_METHODS[min]
CANON_METHODS[maximum] = PWL_METHODS[maximum]
CANON_METHODS[minimum] = PWL_METHODS[minimum]

# Canonicalization of DGPs is a stateful procedure, hence the need for a class.
class DgpCanonMethods(dict):
    def __init__(self, *args, **kwargs) -> None:
        super(DgpCanonMethods, self).__init__(*args, **kwargs)
        self._variables = {}
        self._parameters = {}

    def __contains__(self, key):
        return key in CANON_METHODS

    def __getitem__(self, key):
        if key == Variable:
            return self.variable_canon
        elif key == Parameter:
            return self.parameter_canon
        else:
            return CANON_METHODS[key]

    def variable_canon(self, variable, args):
        del args
        # Swaps out positive variables for unconstrained variables.
        if variable in self._variables:
            return self._variables[variable], []
        else:
            log_variable =  Variable(variable.shape, var_id=variable.id)
            self._variables[variable] = log_variable
            return log_variable, []

    def parameter_canon(self, parameter, args):
        del args
        # Swaps out positive parameters for unconstrained variables.
        if parameter in self._parameters:
            return self._parameters[parameter], []
        else:
            # DGP -> DCP transformation: create log-space parameter.
            # For DGP problem `p_dgp`, we canonicalize to DCP by transforming
            # to log-space: `log(p_dgp)` becomes `p_dcp`.
            #
            # DPP support: Create the log-parameter structure WITHOUT requiring
            # an initial value. This allows get_problem_data(gp=True) to work
            # with uninitialized parameters (issue #3004).
            log_parameter = Parameter(parameter.shape, name=parameter.name())
            if parameter.value is not None:
                log_parameter.value = np.log(parameter.value)
            self._parameters[parameter] = log_parameter
            return log_parameter, []



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Tuple

import numpy as np

import cvxpy.interface as intf
import cvxpy.lin_ops.lin_op as lo
import cvxpy.lin_ops.lin_utils as lu
import cvxpy.utilities as u
from cvxpy.atoms.affine.affine_atom import AffAtom
from cvxpy.constraints.constraint import Constraint
from cvxpy.expressions.constants.parameter import is_param_free
from cvxpy.utilities.warn import CvxpyDeprecationWarning, warn


class conv(AffAtom):
    """ 1D discrete convolution of two vectors.

    The discrete convolution :math:`c` of vectors :math:`a` and :math:`b` of
    lengths :math:`n` and :math:`m`, respectively, is a length-:math:`(n+m-1)`
    vector where

    .. math::

        c_k = \\sum_{i+j=k} a_ib_j, \\quad k=0, \\ldots, n+m-2.

    Parameters
    ----------
    lh_expr : Constant
        A constant 1D vector or a 2D column vector.
    rh_expr : Expression
        A 1D vector or a 2D column vector.
    """

    def __init__(self, lh_expr, rh_expr) -> None:
        warn("conv is deprecated. Use convolve instead.", CvxpyDeprecationWarning)
        super(conv, self).__init__(lh_expr, rh_expr)

    @AffAtom.numpy_numeric
    def numeric(self, values):
        """Convolve the two values.
        """
        flat_values = list(map(intf.from_2D_to_1D, values))
        output = np.convolve(flat_values[0], flat_values[1])
        if values[0].ndim == 2 or values[1].ndim == 2:
            return output[:, None]
        else:
            return output

    def validate_arguments(self) -> None:
        """Checks that both arguments are vectors, and the first is constant.
        """
        if not self.args[0].is_vector() or not self.args[1].is_vector():
            raise ValueError("The arguments to conv must resolve to a 1-d array")
        if not self.args[0].is_constant():
            raise ValueError("The first argument to conv must be constant.")

    def is_atom_convex(self) -> bool:
        """Is the atom convex?
        """
        # TODO support a non-constant first argument.
        if u.scopes.dpp_scope_active():
            # conv is not DPP if any parameters are present.
            c = self.args[0]
            return c.is_constant() and is_param_free(c)
        else:
            return self.args[0].is_constant()

    def is_atom_concave(self) -> bool:
        """Is the atom concave?
        """
        return self.is_atom_convex()

    def shape_from_args(self) -> Tuple[int, int]:
        """The sum of the argument dimensions - 1.
        """
        lh_length = self.args[0].size
        rh_length = self.args[1].size
        output_length = lh_length + rh_length - 1
        if self.args[0].ndim == 2 or self.args[1].ndim == 2:
            return (output_length, 1)
        else:
            return (output_length,)

    def sign_from_args(self) -> Tuple[bool, bool]:
        """Same as times.
        """
        return u.sign.mul_sign(self.args[0], self.args[1])

    def is_incr(self, idx) -> bool:
        """Is the composition non-decreasing in argument idx?
        """
        return self.args[0].is_nonneg()

    def is_decr(self, idx) -> bool:
        """Is the composition non-increasing in argument idx?
        """
        return self.args[0].is_nonpos()

    def graph_implementation(
        self, arg_objs, shape: Tuple[int, ...], data=None
    ) -> Tuple[lo.LinOp, List[Constraint]]:
        """Convolve two vectors.

        Parameters
        ----------
        arg_objs : list
            LinExpr for each argument.
        shape : tuple
            The shape of the resulting expression.
        data :
            Additional data required by the atom.

        Returns
        -------
        tuple
            (LinOp for objective, list of constraints)
        """
        return (lu.conv(arg_objs[0], arg_objs[1], shape), [])


class convolve(AffAtom):
    """ 1D discrete convolution of two vectors.

    The discrete convolution :math:`c` of vectors :math:`a` and :math:`b` of
    lengths :math:`n` and :math:`m`, respectively, is a length-:math:`(n+m-1)`
    vector where

    .. math::

        c_k = \\sum_{i+j=k} a_ib_j, \\quad k=0, \\ldots, n+m-2.

    Matches numpy.convolve

    Parameters
    ----------
    lh_expr : Constant
        A constant scalar or 1D vector.
    rh_expr : Expression
        A scalar or 1D vector.
    """
    # TODO work with right hand constant.
    # TODO(akshayka): make DGP-compatible

    @AffAtom.numpy_numeric
    def numeric(self, values):
        """Convolve the two values.
        """
        return np.convolve(values[0], values[1])

    def validate_arguments(self) -> None:
        """Checks that both arguments are vectors, and the first is constant.
        """
        if not self.args[0].ndim <= 1 or not self.args[1].ndim <= 1:
            raise ValueError("The arguments to conv must be scalar or 1D.")
        if not self.args[0].is_constant():
            raise ValueError("The first argument to conv must be constant.")

    def is_atom_convex(self) -> bool:
        """Is the atom convex?
        """
        # TODO support a non-constant first argument.
        if u.scopes.dpp_scope_active():
            # conv is not DPP if any parameters are present.
            c = self.args[0]
            return c.is_constant() and is_param_free(c)
        else:
            return self.args[0].is_constant()

    def is_atom_concave(self) -> bool:
        """Is the atom concave?
        """
        return self.is_atom_convex()

    def shape_from_args(self) -> Tuple[int, int]:
        """The sum of the argument dimensions - 1.
        """
        lh_length = self.args[0].size
        rh_length = self.args[1].size
        return (lh_length + rh_length - 1,)

    def sign_from_args(self) -> Tuple[bool, bool]:
        """Same as times.
        """
        return u.sign.mul_sign(self.args[0], self.args[1])

    def is_incr(self, idx) -> bool:
        """Is the composition non-decreasing in argument idx?
        """
        return self.args[0].is_nonneg()

    def is_decr(self, idx) -> bool:
        """Is the composition non-increasing in argument idx?
        """
        return self.args[0].is_nonpos()

    def graph_implementation(
        self, arg_objs, shape: Tuple[int, ...], data=None
    ) -> Tuple[lo.LinOp, List[Constraint]]:
        """Convolve two vectors.

        Parameters
        ----------
        arg_objs : list
            LinExpr for each argument.
        shape : tuple
            The shape of the resulting expression.
        data :
            Additional data required by the atom.

        Returns
        -------
        tuple
            (LinOp for objective, list of constraints)
        """
        return (lu.conv(arg_objs[0], arg_objs[1], shape), [])



"""
Copyright, the CVXPY authors
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from cvxpy.error import DCPError, DGPError, SolverError
from cvxpy.problems.objective import Maximize
from cvxpy.reductions.chain import Chain
from cvxpy.reductions.complex2real import complex2real
from cvxpy.reductions.cvx_attr2constr import CvxAttr2Constr
from cvxpy.reductions.dcp2cone.dcp2cone import Dcp2Cone
from cvxpy.reductions.dgp2dcp.dgp2dcp import Dgp2Dcp
from cvxpy.reductions.flip_objective import FlipObjective
from cvxpy.utilities.debug_tools import build_non_disciplined_error_msg


def construct_intermediate_chain(problem, candidates, gp: bool = False):
    """
    Builds a chain that rewrites a problem into an intermediate
    representation suitable for numeric reductions.

    Parameters
    ----------
    problem : Problem
        The problem for which to build a chain.
    candidates : dict
        Dictionary of candidate solvers divided in qp_solvers
        and conic_solvers.
    gp : bool
        If True, the problem is parsed as a Disciplined Geometric Program
        instead of as a Disciplined Convex Program.

    Returns
    -------
    Chain
        A Chain that can be used to convert the problem to an intermediate form.

    Raises
    ------
    DCPError
        Raised if the problem is not DCP and `gp` is False.
    DGPError
        Raised if the problem is not DGP and `gp` is True.
    """

    reductions = []
    if len(problem.variables()) == 0:
        return Chain(reductions=reductions)
    # TODO Handle boolean constraints.
    if complex2real.accepts(problem):
        reductions += [complex2real.Complex2Real()]
    if gp:
        reductions += [Dgp2Dcp()]

    if not gp and not problem.is_dcp():
        append = build_non_disciplined_error_msg(problem, 'DCP')
        if problem.is_dgp():
            append += ("\nHowever, the problem does follow DGP rules. "
                       "Consider calling solve() with `gp=True`.")
        elif problem.is_dqcp():
            append += ("\nHowever, the problem does follow DQCP rules. "
                       "Consider calling solve() with `qcp=True`.")
        raise DCPError("Problem does not follow DCP rules. Specifically:\n" + append)

    elif gp and not problem.is_dgp():
        append = build_non_disciplined_error_msg(problem, 'DGP')
        if problem.is_dcp():
            append += ("\nHowever, the problem does follow DCP rules. "
                       "Consider calling solve() with `gp=False`.")
        elif problem.is_dqcp():
            append += ("\nHowever, the problem does follow DQCP rules. "
                       "Consider calling solve() with `qcp=True`.")
        raise DGPError("Problem does not follow DGP rules." + append)

    # Dcp2Cone requires problems to minimize their objectives.
    if type(problem.objective) == Maximize:
        reductions += [FlipObjective()]

    # First, attempt to canonicalize the problem to a linearly constrained QP.
    if candidates['qp_solvers'] and problem.is_qp():
        reductions += [Dcp2Cone(quad_obj=True),
                       CvxAttr2Constr()]
        return Chain(reductions=reductions)

    # Canonicalize it to conic problem.
    if not candidates['conic_solvers']:
        raise SolverError("Problem could not be reduced to a QP, and no "
                          "conic solvers exist among candidate solvers "
                          "(%s)." % candidates)
    reductions += [Dcp2Cone(),
                   CvxAttr2Constr()]
    return Chain(reductions=reductions)



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Tuple

import numpy as np
import scipy.sparse as sp
from numpy.lib.array_utils import normalize_axis_index

from cvxpy.atoms.affine.affine_atom import AffAtom
from cvxpy.atoms.axis_atom import AxisAtom
from cvxpy.expressions.expression import Expression
from cvxpy.utilities.warn import warn


def _sparse_triu_ones(dim: int) -> sp.csc_array:
    """Create a sparse upper triangular matrix of ones.

    This avoids allocating a dense dim x dim matrix.
    Used for cumsum gradient in CVXPY's convention: grad[i,j] = d(out[j])/d(in[i]).
    """
    # Row i has entries at columns i, i+1, ..., dim-1
    # So row 0 has dim entries, row 1 has dim-1, etc.
    rows = np.repeat(np.arange(dim), np.arange(dim, 0, -1))
    cols = np.concatenate([np.arange(i, dim) for i in range(dim)])
    data = np.ones(len(rows))
    return sp.csc_array((data, (rows, cols)), shape=(dim, dim))


class cumsum(AffAtom, AxisAtom):
    """
    Cumulative sum of the elements of an expression.

    Attributes
    ----------
    expr : CVXPY expression
        The expression being summed.
    axis : int, optional
        The axis to sum across. If None, the array is flattened before cumsum.
        Note: NumPy's default is axis=None, while CVXPY defaults to axis=0.
    """

    _reduce_all_axes_to_none = False

    def __init__(self, expr: Expression, axis: None | int = 0) -> None:
        super(cumsum, self).__init__(expr, axis)

    def validate_arguments(self) -> None:
        """Validate axis, but handle 0D arrays specially."""
        if self.args[0].ndim == 0:
            if self.axis is not None:
                warn(
                    "cumsum on 0-dimensional arrays currently returns a scalar, "
                    "but in a future CVXPY version it will return a 1-element "
                    "array to match numpy.cumsum behavior. Additionally, only "
                    "axis=0, axis=-1, or axis=None will be valid for 0D arrays.",
                    FutureWarning
                )
        else:
            super().validate_arguments()

    @AffAtom.numpy_numeric
    def numeric(self, values):
        """
        Returns the cumulative sum of elements of an expression over an axis.
        """
        return np.cumsum(values[0], axis=self.axis)

    def shape_from_args(self) -> Tuple[int, ...]:
        """Flattened if axis=None, otherwise same as input."""
        if self.axis is None:
            return (self.args[0].size,)
        return self.args[0].shape

    def _grad(self, values):
        """Gives the (sub/super)gradient of the atom w.r.t. each argument.

        Matrix expressions are vectorized, so the gradient is a matrix.
        CVXPY convention: grad[i, j] = d(output[j]) / d(input[i]).

        Args:
            values: A list of numeric values for the arguments.

        Returns:
            A list of SciPy CSC sparse matrices or None.
        """
        ndim = len(values[0].shape)
        axis = self.axis

        # Handle axis=None: treat as 1D cumsum over C-order flattened array
        if axis is None:
            dim = values[0].size
            # For cumsum with axis=None:
            # - Input x is vectorized in F-order (CVXPY convention)
            # - cumsum flattens in C-order then computes cumsum
            # - Let x_f = F-order input, x_c = C-order = P @ x_f
            # - y = L @ x_c = L @ P @ x_f (L is lower triangular)
            # - dy/dx_f = L @ P
            # - CVXPY wants grad[i,j] = dy[j]/dx_f[i] = (L @ P).T = P.T @ L.T = P.T @ U
            # where U is upper triangular
            triu = _sparse_triu_ones(dim)
            # Permutation: P @ f_vec = c_vec
            c_order_indices = np.arange(dim).reshape(values[0].shape, order='F').flatten(order='C')
            P = sp.csc_array((np.ones(dim), (np.arange(dim), c_order_indices)), shape=(dim, dim))
            grad = P.T @ triu
            return [sp.csc_array(grad)]

        axis = normalize_axis_index(axis, ndim)
        dim = values[0].shape[axis]

        # Upper triangular matrix for CVXPY gradient convention
        # grad[i, j] = d(cumsum[j])/d(x[i]) = 1 if i <= j
        triu = _sparse_triu_ones(dim)

        # Kronecker product: I_post ⊗ triu ⊗ I_pre
        # This works for all dimensions including 1D and 2D
        pre_size = int(np.prod(values[0].shape[:axis])) if axis > 0 else 1
        post_size = int(np.prod(values[0].shape[axis+1:])) if axis < ndim - 1 else 1

        grad = sp.kron(sp.kron(sp.eye_array(post_size), triu), sp.eye_array(pre_size))
        return [sp.csc_array(grad)]

    def get_data(self):
        """Returns the axis being summed."""
        return [self.axis]



import numpy as np
import scipy.sparse as sp

import cvxpy
from cvxpy.tests.base_test import BaseTest


class TestDgp(BaseTest):
    def test_product(self) -> None:
        x = cvxpy.Variable((), pos=True)
        y = cvxpy.Variable((), pos=True)
        prod = x * y
        self.assertTrue(prod.is_dgp())
        self.assertTrue(prod.is_log_log_convex())
        self.assertTrue(prod.is_log_log_concave())

        prod *= prod
        self.assertTrue(prod.is_dgp())
        self.assertTrue(prod.is_log_log_convex())
        self.assertTrue(prod.is_log_log_concave())

        prod *= 5.0
        self.assertTrue(prod.is_dgp())
        self.assertTrue(prod.is_log_log_convex())
        self.assertTrue(prod.is_log_log_concave())

        prod *= -5.0
        self.assertTrue(not prod.is_dgp())
        self.assertTrue(not prod.is_log_log_convex())
        self.assertTrue(not prod.is_log_log_concave())

    def test_product_with_unconstrained_variables_is_not_dgp(self) -> None:
        x = cvxpy.Variable()
        y = cvxpy.Variable()
        prod = x * y
        self.assertTrue(not prod.is_dgp())
        self.assertTrue(not prod.is_log_log_convex())
        self.assertTrue(not prod.is_log_log_concave())

        z = cvxpy.Variable((), pos=True)
        prod = x * z
        self.assertTrue(not prod.is_dgp())
        self.assertTrue(not prod.is_log_log_convex())
        self.assertTrue(not prod.is_log_log_concave())

    def test_division(self) -> None:
        x = cvxpy.Variable(pos=True)
        y = cvxpy.Variable(pos=True)
        div = x / y

        self.assertTrue(div.is_log_log_affine())

        posynomial = 5.0 * x * y + 1.2 * y * y
        div = x / y
        self.assertTrue(div.is_log_log_affine())

        div = posynomial / (3.0 * x * y ** (-0.1))
        self.assertTrue(div.is_log_log_convex())
        self.assertFalse(div.is_log_log_concave())
        self.assertTrue(div.is_dgp())

        div = posynomial / (3.0 * x + y)
        self.assertFalse(div.is_log_log_convex())
        self.assertFalse(div.is_log_log_concave())
        self.assertFalse(div.is_dgp())

    def test_add(self) -> None:
        x = cvxpy.Variable(pos=True)
        y = cvxpy.Variable(pos=True)
        expr = x + y
        self.assertTrue(expr.is_dgp())
        self.assertTrue(expr.is_log_log_convex())
        self.assertTrue(not expr.is_log_log_concave())

        posynomial = 5.0 * x * y + 1.2 * y * y
        self.assertTrue(posynomial.is_dgp())
        self.assertTrue(posynomial.is_log_log_convex())

    def test_add_with_unconstrained_variables_is_not_dgp(self) -> None:
        x = cvxpy.Variable()
        y = cvxpy.Variable(pos=True)
        expr = x + y
        self.assertTrue(not expr.is_dgp())
        self.assertTrue(not expr.is_log_log_convex())
        self.assertTrue(not expr.is_log_log_concave())

        posynomial = 5.0 * x * y + 1.2 * y * y
        self.assertTrue(not posynomial.is_dgp())
        self.assertTrue(not posynomial.is_log_log_convex())
        self.assertTrue(not posynomial.is_log_log_concave())

    def test_monomials(self) -> None:
        x = cvxpy.Variable(pos=True)
        y = cvxpy.Variable(pos=True)
        z = cvxpy.Variable(pos=True)
        monomial = 5.0 * (x ** 0.1) * y ** (-0.1) * z ** (3)
        self.assertTrue(monomial.is_dgp())
        self.assertTrue(monomial.is_log_log_convex())
        self.assertTrue(monomial.is_log_log_concave())

        monomial *= -1.0
        self.assertTrue(not monomial.is_dgp())
        self.assertTrue(not monomial.is_log_log_convex())
        self.assertTrue(not monomial.is_log_log_concave())

    def test_maximum(self) -> None:
        x = cvxpy.Variable(pos=True)
        y = cvxpy.Variable(pos=True)
        z = cvxpy.Variable(pos=True)
        monomial = 5.0 * (x ** 0.1) * y ** (-0.1) * z ** (3)
        posynomial = 5.0 * x * y + 1.2 * y * y
        another_posynomial = posynomial * posynomial
        expr = cvxpy.maximum(monomial, posynomial, another_posynomial)
        self.assertTrue(expr.is_dgp())
        self.assertTrue(expr.is_log_log_convex())
        self.assertTrue(not expr.is_log_log_concave())

        expr = posynomial * expr
        self.assertTrue(expr.is_dgp())
        self.assertTrue(expr.is_log_log_convex())
        self.assertTrue(not expr.is_log_log_concave())

        expr = posynomial * expr + expr
        self.assertTrue(expr.is_dgp())
        self.assertTrue(expr.is_log_log_convex())

    def test_minimum(self) -> None:
        x = cvxpy.Variable(pos=True)
        y = cvxpy.Variable(pos=True)
        z = cvxpy.Variable(pos=True)
        monomial = 5.0 * (x ** 0.1) * y ** (-0.1) * z ** (3)
        posynomial = 5.0 * x * y + 1.2 * y * y
        another_posynomial = posynomial * posynomial
        expr = cvxpy.minimum(monomial, 1 / posynomial, 1 / another_posynomial)
        self.assertTrue(expr.is_dgp())
        self.assertTrue(not expr.is_log_log_convex())
        self.assertTrue(expr.is_log_log_concave())

        expr = (1 / posynomial) * expr
        self.assertTrue(expr.is_dgp())
        self.assertTrue(not expr.is_log_log_convex())
        self.assertTrue(expr.is_log_log_concave())

        expr = expr ** 2
        self.assertTrue(expr.is_dgp())
        self.assertTrue(not expr.is_log_log_convex())
        self.assertTrue(expr.is_log_log_concave())

    def test_constant(self) -> None:
        x = cvxpy.Constant(1.0)
        self.assertTrue(x.is_dgp())
        self.assertFalse((-1.0*x).is_dgp())

    def test_geo_mean(self) -> None:
        x = cvxpy.Variable(3, pos=True)
        p = [1, 2, 0.5]
        geo_mean = cvxpy.geo_mean(x, p)
        self.assertTrue(geo_mean.is_dgp())
        self.assertTrue(geo_mean.is_log_log_affine())
        self.assertTrue(geo_mean.is_log_log_convex())
        self.assertTrue(geo_mean.is_log_log_concave())

    def test_geo_mean_scalar1(self) -> None:
        x = cvxpy.Variable(1, pos=True)
        p = np.array([2])
        geo_mean = cvxpy.geo_mean(x, p)
        self.assertTrue(geo_mean.is_dgp())
        prob = cvxpy.Problem(
            cvxpy.Maximize(geo_mean),
            [x == 2],
        )
        prob.solve()
        self.assertEqual(prob.value, 2)

    def test_geo_mean_scalar2(self) -> None:
        x = cvxpy.Variable(pos=True)
        p = np.array([2])
        geo_mean = cvxpy.geo_mean(x, p)
        self.assertTrue(geo_mean.is_dgp())

    def test_inv_prod(self) -> None:
        x = cvxpy.Variable(2)
        # # test inv_prod with scalar value
        prob1 = cvxpy.Problem(
            cvxpy.Minimize(cvxpy.inv_prod(x[0]) + cvxpy.inv_prod(x[:2])),
            [cvxpy.sum(x) == 2],
        )
        prob1.solve()

        # compare inv_prod with inv_pos
        prob2 = cvxpy.Problem(
            cvxpy.Minimize(cvxpy.inv_prod(x[:1]) + cvxpy.inv_prod(x[:2])),
            [cvxpy.sum(x) == 2],
        )
        prob2.solve()

        prob3 = cvxpy.Problem(
            cvxpy.Minimize(cvxpy.inv_pos(x[0]) + cvxpy.inv_prod(x[:2])),
            [cvxpy.sum(x) == 2],
        )
        prob3.solve()
        self.assertAlmostEqual(prob2.value, prob3.value, 4)

    def test_builtin_sum(self) -> None:
        x = cvxpy.Variable(2, pos=True)
        self.assertTrue(sum(x).is_log_log_convex())

    def test_gmatmul(self) -> None:
        x = cvxpy.Variable(2, pos=True)
        A = cvxpy.Variable((2, 2))
        with self.assertRaises(Exception) as cm:
            cvxpy.gmatmul(A, x)
        self.assertTrue(str(cm.exception) ==
                        "gmatmul(A, X) requires that A be constant.")

        x = cvxpy.Variable(2)
        A = np.ones((4, 2))
        with self.assertRaises(Exception) as cm:
            cvxpy.gmatmul(A, x)
        self.assertTrue(str(cm.exception) ==
                        "gmatmul(A, X) requires that X be positive.")

        x = cvxpy.Variable(3, pos=True)
        A = np.ones((4, 3))
        gmatmul = cvxpy.gmatmul(A, x)
        self.assertTrue(gmatmul.is_dgp())
        self.assertTrue(gmatmul.is_log_log_affine())
        self.assertTrue(gmatmul.is_log_log_convex())
        self.assertTrue(gmatmul.is_log_log_concave())
        self.assertTrue(gmatmul.is_nonneg())
        self.assertTrue(gmatmul.is_incr(0))
        self.assertTrue(cvxpy.gmatmul(-A, x).is_decr(0))

        x = cvxpy.Variable((2, 3), pos=True)
        A = np.array([[2., -1.], [0., 3.]])
        gmatmul = cvxpy.gmatmul(A, x)
        self.assertTrue(gmatmul.is_dgp())
        self.assertTrue(gmatmul.is_log_log_affine())
        self.assertTrue(gmatmul.is_log_log_convex())
        self.assertTrue(gmatmul.is_log_log_concave())
        self.assertFalse(gmatmul.is_incr(0))
        self.assertFalse(gmatmul.is_decr(0))

    def test_power_sign(self) -> None:
        x = cvxpy.Variable(pos=True)
        self.assertTrue((x**1).is_nonneg())
        self.assertFalse((x**1).is_nonpos())

    def test_sparse_constant_not_allowed(self) -> None:
        sparse_matrix = cvxpy.Constant(sp.csc_array(np.array([[1.0, 2.0]])))
        self.assertFalse(sparse_matrix.is_log_log_constant())

    def test_sparse_variable_not_dgp(self) -> None:
        """Test that sparse/diag + pos/neg variables are rejected at construction."""
        rows = np.array([0, 1, 2])
        cols = np.array([0, 1, 2])

        # pos + sparsity
        with self.assertRaises(ValueError):
            cvxpy.Variable((3, 3), sparsity=(rows, cols), pos=True)

        # neg + sparsity
        with self.assertRaises(ValueError):
            cvxpy.Variable((3, 3), sparsity=(rows, cols), neg=True)

        # pos + diag
        with self.assertRaises(ValueError):
            cvxpy.Variable(3, diag=True, pos=True)

        # neg + diag
        with self.assertRaises(ValueError):
            cvxpy.Variable(3, diag=True, neg=True)

    def test_pnorm_scalar(self) -> None:
        """Regression test: scalar DGP pnorm must canonicalize x, not the exponent p."""
        # pnorm of a single positive scalar equals that scalar,
        # so the minimum subject to x >= lb must be lb.
        x = cvxpy.Variable(pos=True)
        prob = cvxpy.Problem(cvxpy.Minimize(cvxpy.pnorm(x, p=2)), [x >= 2.5])
        prob.solve(gp=True)
        self.assertEqual(prob.status, cvxpy.OPTIMAL)
        np.testing.assert_allclose(x.value, 2.5, atol=1e-4)

        x = cvxpy.Variable(pos=True)
        prob = cvxpy.Problem(cvxpy.Minimize(cvxpy.pnorm(x, p=3)), [x >= 4.0])
        prob.solve(gp=True)
        self.assertEqual(prob.status, cvxpy.OPTIMAL)
        np.testing.assert_allclose(x.value, 4.0, atol=1e-4)

    def test_dgp_sum_3d_axis(self) -> None:
        """Test DGP sum canonicalizer on 3D with 2+ keep dimensions.

        Uses non-uniform weights on sum(x, axis=1) in the objective so
        element ordering matters.  A reshape-order bug would transpose
        the canonical expression, multiplying wrong weights with wrong
        sums and producing a different objective value.
        """
        # x shape (2, 3, 4), reduce axis=1 → output shape (2, 4)
        x = cvxpy.Variable((2, 3, 4), pos=True)
        c = np.random.RandomState(42).uniform(0.5, 2.0, (2, 3, 4))
        s = cvxpy.sum(x, axis=1)  # shape (2, 4)

        # Non-uniform weights — transposition changes which weight
        # multiplies which sum element, altering the optimal value.
        w = np.array([[1, 2, 3, 4],
                      [5, 6, 7, 8]], dtype=float)

        # Minimize weighted sum; at optimum x == c (tight lower bound).
        prob = cvxpy.Problem(cvxpy.Minimize(cvxpy.sum(cvxpy.multiply(w, s))),
                             [x >= c])
        prob.solve(gp=True)

        expected_s = np.sum(c, axis=1)  # shape (2, 4)
        expected_obj = float(np.sum(w * expected_s))
        self.assertAlmostEqual(prob.value, expected_obj, places=2)

    def test_dgp_pnorm_3d_axis(self) -> None:
        """Test DGP pnorm canonicalizer on 3D with 2+ keep dimensions.

        Constrains x == c so the pnorm result is deterministic, then
        checks that the objective value matches the expected pnorm.
        Because pnorm(x, axis=1) appears in the objective (not just
        as a .value query), element ordering matters.
        """
        x = cvxpy.Variable((2, 3, 4), pos=True)
        c = np.random.RandomState(43).uniform(0.5, 2.0, (2, 3, 4))

        # axis=1, p=2: reduce middle axis, output shape (2, 4)
        pn = cvxpy.pnorm(x, p=2, axis=1)
        prob = cvxpy.Problem(cvxpy.Minimize(cvxpy.sum(pn)), [x == c])
        prob.solve(gp=True)

        expected = np.linalg.norm(c, ord=2, axis=1)  # shape (2, 4)
        self.assertAlmostEqual(prob.value, float(np.sum(expected)), places=2)

    def test_dgp_infeasible_no_crash(self) -> None:
        """Infeasible DGP problem should not crash in invert()."""
        x = cvxpy.Variable(pos=True)
        prob = cvxpy.Problem(cvxpy.Minimize(x), [x >= 2, x <= 1])
        prob.solve(solver=cvxpy.CLARABEL, gp=True)
        self.assertIn(prob.status, [cvxpy.INFEASIBLE, cvxpy.INFEASIBLE_INACCURATE])

    def test_pnorm_negative_p_dgp(self) -> None:
        """pnorm with p < 0 is log-log concave, so Maximize should be DGP."""
        x = cvxpy.Variable(3, pos=True)
        # Maximize pnorm(x, p=-1) s.t. x <= 2, x >= 0.5
        # Optimal: all x_i = 2, pnorm = (3 * 2^(-1))^(-1) = 2/3
        prob = cvxpy.Problem(
            cvxpy.Maximize(cvxpy.pnorm(x, p=-1)),
            [x <= 2, x >= 0.5],
        )
        self.assertTrue(prob.is_dgp())
        prob.solve(gp=True, solver=cvxpy.SCS)
        self.assertEqual(prob.status, cvxpy.OPTIMAL)
        self.assertAlmostEqual(prob.value, 2.0 / 3.0, places=3)

        # Minimize pnorm(x, p=-1) should NOT be DGP.
        prob2 = cvxpy.Problem(
            cvxpy.Minimize(cvxpy.pnorm(x, p=-1)),
            [x <= 2, x >= 0.5],
        )
        self.assertFalse(prob2.is_dgp())



from cvxpy.constraints.finite_set import FiniteSet


def finite_set_canon(expr, args):
    ineq_form, id = expr.get_data()
    # Log applied already to set of values (i.e., args[1]).
    return FiniteSet(args[0], args[1], ineq_form, id), []



def unique_list(duplicates_list):
    """
    Return unique list preserving the order.
    https://stackoverflow.com/a/480227
    """
    used = set()
    unique = [x for x in duplicates_list if not (x in used or used.add(x))]
    return unique




# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '1.8.2'
version = '1.8.2'
full_version = '1.8.2'
git_revision = '76b0527'
commit_count = '255'
release = True
if not release:
    version = full_version



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import contextlib
import threading
from typing import Generator

_thread_local = threading.local()


@contextlib.contextmanager
def dpp_scope() -> Generator[None, None, None]:
    """Context manager for DPP curvature analysis

    When this scope is active, parameters are affine, not constant. The
    argument For example, if `param` is a Parameter, then

    ```
        with dpp_scope():
            print("param is constant: ", param.is_constant())
            print("param is affine: ", param.is_affine())
    ```

    would print

        param is constant: False
        param is affine: True
    """
    prev_state = getattr(_thread_local, 'dpp_scope_active', False)
    _thread_local.dpp_scope_active = True
    yield
    _thread_local.dpp_scope_active = prev_state


def dpp_scope_active() -> bool:
    """Returns True if a `dpp_scope` is active. """
    return getattr(_thread_local, 'dpp_scope_active', False)



"""
Copyright 2013 Steven Diamond, Eric Chu

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math

import numpy as np

import cvxpy as cp
from cvxpy.tests.base_test import BaseTest


class TestNonlinearAtoms(BaseTest):
    """ Unit tests for the nonlinear atoms module. """

    def setUp(self) -> None:
        self.x = cp.Variable(2, name='x')
        self.y = cp.Variable(2, name='y')

        self.A = cp.Variable((2, 2), name='A')
        self.B = cp.Variable((2, 2), name='B')
        self.C = cp.Variable((3, 2), name='C')

    def test_log_problem(self) -> None:
        # Log in objective.
        obj = cp.Maximize(cp.sum(cp.log(self.x)))
        constr = [self.x <= [1, math.e]]
        p = cp.Problem(obj, constr)
        result = p.solve(solver=cp.CLARABEL)
        self.assertAlmostEqual(result, 1)
        self.assertItemsAlmostEqual(self.x.value, [1, math.e])

        # Log in constraint.
        obj = cp.Minimize(cp.sum(self.x))
        constr = [cp.log(self.x) >= 0, self.x <= [1, 1]]
        p = cp.Problem(obj, constr)
        result = p.solve(solver=cp.CLARABEL)
        self.assertAlmostEqual(result, 2)
        self.assertItemsAlmostEqual(self.x.value, [1, 1])

        # Index into log.
        obj = cp.Maximize(cp.log(self.x)[1])
        constr = [self.x <= [1, math.e]]
        p = cp.Problem(obj, constr)
        result = p.solve(solver=cp.CLARABEL)
        self.assertAlmostEqual(result, 1)

        # Scalar log.
        obj = cp.Maximize(cp.log(self.x[1]))
        constr = [self.x <= [1, math.e]]
        p = cp.Problem(obj, constr)
        result = p.solve(solver=cp.CLARABEL)
        self.assertAlmostEqual(result, 1)

    def test_entr(self) -> None:
        """Test the entr atom.
        """
        self.assertEqual(cp.entr(0).value, 0)
        assert np.isneginf(cp.entr(-1).value)

    def test_kl_div(self) -> None:
        """Test a problem with kl_div.
        """
        kK = 50
        kSeed = 10

        prng = np.random.RandomState(kSeed)
        # Generate a random reference distribution
        npSPriors = prng.uniform(0.0, 1.0, kK)
        npSPriors = npSPriors / sum(npSPriors)

        # Reference distribution
        p_refProb = cp.Parameter(kK, nonneg=True)
        # Distribution to be estimated
        v_prob = cp.Variable(kK)
        objkl = cp.sum(cp.kl_div(v_prob, p_refProb))

        constrs = [cp.sum(v_prob) == 1]
        klprob = cp.Problem(cp.Minimize(objkl), constrs)
        p_refProb.value = npSPriors
        klprob.solve(solver=cp.SCS)
        self.assertItemsAlmostEqual(v_prob.value, npSPriors, places=3)
        klprob.solve(solver=cp.CLARABEL)
        self.assertItemsAlmostEqual(v_prob.value, npSPriors, places=3)

    def test_rel_entr(self) -> None:
        """Test a problem with rel_entr.
        """
        kK = 50
        kSeed = 10

        prng = np.random.RandomState(kSeed)
        # Generate a random reference distribution
        npSPriors = prng.uniform(0.0, 1.0, kK)
        npSPriors = npSPriors / sum(npSPriors)

        # Reference distribution
        p_refProb = cp.Parameter(kK, nonneg=True)
        # Distribution to be estimated
        v_prob = cp.Variable(kK)
        obj_rel_entr = cp.sum(cp.rel_entr(v_prob, p_refProb))

        constrs = [cp.sum(v_prob) == 1]
        rel_entr_prob = cp.Problem(cp.Minimize(obj_rel_entr), constrs)
        p_refProb.value = npSPriors
        rel_entr_prob.solve(solver=cp.SCS)
        self.assertItemsAlmostEqual(v_prob.value, npSPriors, places=3)
        rel_entr_prob.solve(solver=cp.CLARABEL)
        self.assertItemsAlmostEqual(v_prob.value, npSPriors, places=3)

    def test_difference_kl_div_rel_entr(self) -> None:
        """A test showing the difference between kl_div and rel_entr
        """
        x = cp.Variable()
        y = cp.Variable()

        kl_div_prob = cp.Problem(cp.Minimize(cp.kl_div(x, y)), constraints=[x + y <= 1])
        kl_div_prob.solve(solver=cp.CLARABEL)
        self.assertItemsAlmostEqual(x.value, y.value, places=3)
        self.assertItemsAlmostEqual(kl_div_prob.value, 0)

        rel_entr_prob = cp.Problem(cp.Minimize(cp.rel_entr(x, y)), constraints=[x + y <= 1])
        rel_entr_prob.solve(solver=cp.CLARABEL)

        """
        Reference solution computed by passing the following command to Wolfram Alpha:
        minimize x*log(x/y) subject to {x + y <= 1, 0 <= x, 0 <= y}
        """
        self.assertItemsAlmostEqual(x.value, 0.2178117, places=4)
        self.assertItemsAlmostEqual(y.value, 0.7821882, places=4)
        self.assertItemsAlmostEqual(rel_entr_prob.value, -0.278464)

    def test_entr_prob(self) -> None:
        """Test a problem with entr.
        """
        for n in [5, 10, 25]:
            x = cp.Variable(n)
            obj = cp.Maximize(cp.sum(cp.entr(x)))
            p = cp.Problem(obj, [cp.sum(x) == 1])
            p.solve(solver=cp.CLARABEL)
            self.assertItemsAlmostEqual(x.value, n*[1./n], places=3)
            p.solve(solver=cp.SCS)
            self.assertItemsAlmostEqual(x.value, n*[1./n], places=3)

    def test_exp(self) -> None:
        """Test a problem with exp.
        """
        for n in [5, 10, 25]:
            x = cp.Variable(n)
            obj = cp.Minimize(cp.sum(cp.exp(x)))
            p = cp.Problem(obj, [cp.sum(x) == 1])
            p.solve(solver=cp.SCS)
            self.assertItemsAlmostEqual(x.value, n*[1./n], places=3)
            p.solve(solver=cp.CLARABEL)
            self.assertItemsAlmostEqual(x.value, n*[1./n], places=4)

    def test_log(self) -> None:
        """Test a problem with log.
        """
        for n in [5, 10, 25]:
            x = cp.Variable(n)
            obj = cp.Maximize(cp.sum(cp.log(x)))
            p = cp.Problem(obj, [cp.sum(x) == 1])
            p.solve(solver=cp.CLARABEL)
            self.assertItemsAlmostEqual(x.value, n*[1./n])
            p.solve(solver=cp.SCS)
            self.assertItemsAlmostEqual(x.value, n*[1./n], places=2)



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from cvxpy.atoms.quad_over_lin import quad_over_lin


def sum_squares(
    expr,
    axis: None | int | tuple[int, ...] = None,
    keepdims: bool = False
):
    """The sum of the squares of the entries.

    Parameters
    ----------
    expr : Expression
        The expression to take the sum of squares of.
    axis : int or tuple of int, optional
        The axis or axes along which to compute the sum of squares.
        If None (default), sums over all elements and returns a scalar.
        If an int, sums over that axis.
        If a tuple, sums over multiple axes.
    keepdims : bool, optional
        If True, the reduced axes are retained as dimensions with size 1.
        Default is False.

    Returns
    -------
    Expression
        An expression representing the sum of squares.
        Scalar if axis is None, otherwise a vector/array (or with keepdims=True,
        retains shape with reduced dimensions as 1).
    """
    return quad_over_lin(expr, 1, axis=axis, keepdims=keepdims)
